# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
# This file is intentionally left blank to mark the directory as a package.
from __future__ import annotations



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import io
import logging
import os
import pathlib
import tarfile
import tempfile
import time
import webbrowser
from collections.abc import Sequence
from enum import IntEnum
from typing import BinaryIO

import hid
import python_minifier
import requests
import semver
from mutenix.hid_commands import VersionInfo
from tqdm import tqdm

_logger = logging.getLogger(__name__)


def check_for_device_update(
    device: hid.device,
    device_version: VersionInfo,
    proxy: str | None = None,
):
    if proxy:
        proxies = {"https": proxy}
    else:
        proxies = {}
    try:
        result = requests.get(
            "https://api.github.com/repos/mutenix-org/firmware-macroboard/releases/latest",
            timeout=4,
            proxies=proxies,
        )
        if result.status_code != 200:
            _logger.error(
                "Failed to fetch latest release info, status code: %s",
                result.status_code,
            )
            return

        releases = result.json()
        latest_version = releases.get("tag_name", "v0.0.0")[1:]
        _logger.debug("Latest version: %s", latest_version)
        online_version = semver.Version.parse(latest_version)
        local_version = semver.Version.parse(device_version.version)
        if online_version.compare(local_version) <= 0:
            _logger.info("Device is up to date")
            return

        print("Device update available, starting update, please be patient")
        assets = releases.get("assets", [])
        for asset in assets:
            if asset.get("name") == f"v{latest_version}.tar.gz":
                update_url = asset.get("browser_download_url")
                result = requests.get(update_url)
                result.raise_for_status()
                perform_upgrade_with_file(device, io.BytesIO(result.content))
                return
    except requests.RequestException as e:
        _logger.error("Failed to check for device update availability %s", e)


HEADER_SIZE = 8
MAX_CHUNK_SIZE = 60 - HEADER_SIZE
DATA_TRANSFER_SLEEP_TIME = 1
STATE_CHANGE_SLEEP_TIME = 0.5
WAIT_FOR_REQUESTS_SLEEP_TIME = STATE_CHANGE_SLEEP_TIME
HID_REPORT_ID_COMMUNICATION = 1
HID_REPORT_ID_TRANSFER = 2

HID_COMMAND_PREPARE_UPDATE = 0xE0
HID_COMMAND_RESET = 0xE1


def perform_upgrade_with_file(device: hid.device, file_stream: BinaryIO):
    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdir = pathlib.Path(tmpdirname)
        with tarfile.open(fileobj=file_stream, mode="r:gz") as tar:
            tar.extractall(path=tmpdirname)
        files = list(
            map(
                lambda x: tmpdir / x,
                filter(
                    lambda x: (x.endswith(".py") or x.endswith(".delete"))
                    and not x.startswith("."),
                    os.listdir(tmpdirname),
                ),
            ),
        )
        _logger.debug("Updateing device with files: %s", files)
        perform_hid_upgrade(device, files)
        _logger.info("Successfully updated device firmware")


class UpdateError:
    def __init__(self, data: bytes):
        self.info = ""
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        length = max(int.from_bytes(data[2:3], "little"), 33)
        self.info = data[3 : 3 + length].decode("utf-8")
        _logger.info("Error received: %s", self)

    @property
    def is_valid(self):
        return self.identifier == "ER"

    def __str__(self):
        if self.is_valid:
            return f"Error: {self.info}"
        return "Invalid Request"


class ChunkAck:
    def __init__(self, data: bytes):
        self.id = 0
        self.package = 0
        self.type_ = 0
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        self.id = int.from_bytes(data[2:4], "little")
        self.package = int.from_bytes(data[4:6], "little")
        self.type_ = int.from_bytes(data[6:7], "little")
        _logger.info("Ack: %s", self)

    @property
    def is_valid(self):
        return self.identifier == "AK"

    def __str__(self):
        if self.is_valid:
            return f"File: {self.id}, Type: {self.type_}, Package: {self.package}"
        return "Invalid Request"


class LogMessage:
    def __init__(self, data: bytes):
        self.message = ""
        self.parse(data)

    def parse(self, data: bytes):
        self.identifier = data[:2].decode("utf-8")
        if not self.is_valid:
            return
        self.level = "debug" if self.identifier == "LD" else "error"
        self.message = data[2 : data.index(0)].decode("utf-8")

    @property
    def is_valid(self):
        return self.identifier in ("LE", "LD")

    def __str__(self):
        if self.is_valid:
            return f"{self.level}: {self.message}"
        return "Invalid Request"


def parse_hid_update_message(data: bytes) -> ChunkAck | UpdateError | LogMessage | None:
    if len(data) < 2:
        return None
    val = data[:2].decode("utf-8")
    match val:
        case "AK":
            return ChunkAck(data)
        case "ER":
            return UpdateError(data)
        case "LD", "LE":
            return LogMessage(data)
    return None


class ChunkType(IntEnum):
    FILE_START = 1
    FILE_CHUNK = 2
    FILE_END = 3
    COMPLETE = 4
    FILE_DELETE = 5


class Chunk:
    def __init__(self, type_: int, id: int, package: int = 0, total_packages: int = 0):
        self.type_ = type_
        self.id = id
        self.package = package
        self.total_packages = total_packages
        self._acked = False
        self.content = b""

    def packet(self):  # pragma: no cover
        return (
            self._base_packet()
            + self.content
            + b"\0" * (MAX_CHUNK_SIZE - len(self.content))
        )

    def _base_packet(self):
        return (
            int(self.type_).to_bytes(2, "little")
            + self.id.to_bytes(2, "little")
            + self.total_packages.to_bytes(2, "little")
            + self.package.to_bytes(2, "little")
        )

    @property
    def acked(self):
        return self._acked


class FileChunk(Chunk):
    def __init__(self, id: int, package: int, total_packages: int, content: bytes):
        super().__init__(ChunkType.FILE_CHUNK, id, package, total_packages)
        self.content = content


class FileStart(Chunk):
    def __init__(
        self,
        id: int,
        package: int,
        total_packages: int,
        filename: str,
        filesize: int,
    ):
        super().__init__(ChunkType.FILE_START, id, package, total_packages)
        self.content = (
            bytes((len(filename),))
            + filename.encode("utf-8")
            + bytes((2,))
            + filesize.to_bytes(2, "little")
        )


class FileEnd(Chunk):
    def __init__(self, id: int):
        super().__init__(ChunkType.FILE_END, id)


class FileDelete(Chunk):
    def __init__(
        self,
        id: int,
        filename: str,
    ):
        super().__init__(ChunkType.FILE_DELETE, id)
        self.content = bytes((len(filename),)) + filename.encode("utf-8")


class Completed(Chunk):
    def __init__(self):
        super().__init__(ChunkType.COMPLETE, 0)


class TransferFile:
    def __init__(self, id, filename: str | pathlib.Path):
        self.id = id
        if isinstance(filename, str):
            file = pathlib.Path(filename)
        else:
            file = filename
        self.filename = file.name
        self.packages_sent: list[int] = []
        self._chunks: list[Chunk] = []
        if self.filename.endswith(".delete"):
            self.filename = self.filename[:-7]
            self._chunks = [FileDelete(self.id, self.filename)]
            return

        with open(file, "r") as f:
            if file.suffix == ".py":
                self.content = python_minifier.minify(
                    f.read(),
                    remove_annotations=True,
                    rename_globals=False,
                ).encode("utf-8")
            else:
                self.content = f.read()
        self.size = len(self.content)
        self.make_chunks()
        _logger.debug("File %s has %s chunks", self.filename, len(self._chunks))

    def make_chunks(self):
        total_packages = self.size // MAX_CHUNK_SIZE
        self._chunks.append(
            FileStart(self.id, 0, total_packages, self.filename, self.size),
        )
        for i in range(0, self.size, MAX_CHUNK_SIZE):
            self._chunks.append(
                FileChunk(
                    self.id,
                    i // MAX_CHUNK_SIZE,
                    total_packages,
                    self.content[i : i + MAX_CHUNK_SIZE],
                ),
            )
        self._chunks.append(FileEnd(self.id))

    def get_next_chunk(self) -> Chunk | None:
        if self.is_complete():
            return None
        return next((chunk for chunk in self._chunks if not chunk.acked))

    def ack_chunk(self, chunk: ChunkAck):
        if chunk.id != self.id:  # pragma: no cover
            return

        acked_chunk = next(
            filter(
                lambda x: x.type_ == chunk.type_ and x.package == chunk.package,
                self._chunks,
            ),
            None,
        )
        if not acked_chunk:  # pragma: no cover
            _logger.warning("No chunk found for ack")
            return
        acked_chunk._acked = True
        _logger.info("Acked chunk %s", chunk.package)

    @property
    def chunks(self):
        return len(self._chunks)

    def is_complete(self):
        return all(map(lambda x: x.acked, self._chunks))


def send_hid_command(device: hid.device, command: int):
    device.write([HID_REPORT_ID_COMMUNICATION, command] + [0] * 7)


def perform_hid_upgrade(device: hid.device, files: Sequence[str | pathlib.Path]):
    _logger.debug("Opening device for update")
    _logger.debug("Sending prepare update")
    send_hid_command(device, HID_COMMAND_PREPARE_UPDATE)
    time.sleep(STATE_CHANGE_SLEEP_TIME)

    transfer_files = [TransferFile(i, file) for i, file in enumerate(files)]

    _logger.debug("Preparing to send %s files", len(transfer_files))
    cancelled = False

    for i, file in enumerate(transfer_files, 1):
        if cancelled:
            break
        fileprogress = tqdm(
            total=file.chunks,
            desc=f"Sending file {file.filename:25} {i:2}/{len(transfer_files)}",
        )
        while True:
            received = device.read(100, 1000)
            if len(received) > 0:
                rcvd = parse_hid_update_message(bytes(received[1:]))

                if isinstance(rcvd, ChunkAck):
                    ack_file = next(
                        (f for f in transfer_files if f.id == rcvd.id),
                        None,
                    )
                    if not ack_file:
                        _logger.warning("No file id found for ack")
                        continue
                    fileprogress.update(1)
                    ack_file.ack_chunk(rcvd)
                elif isinstance(rcvd, UpdateError):
                    print("Error received from device: ", rcvd)
                    _logger.error("Error received from device: %s", rcvd)
                    cancelled = True
                    break
                elif isinstance(rcvd, LogMessage):
                    print(rcvd)

            chunk = file.get_next_chunk()
            if not chunk:
                fileprogress.close()
                break
            _logger.debug(
                "Sending chunk (%s...) of file %s",
                chunk.packet()[:10],
                file.filename,
            )
            cnk = bytes((HID_REPORT_ID_TRANSFER,)) + chunk.packet()
            device.write(cnk)

    time.sleep(STATE_CHANGE_SLEEP_TIME)
    device.write(bytes((HID_REPORT_ID_TRANSFER,)) + Completed().packet())
    time.sleep(STATE_CHANGE_SLEEP_TIME)
    print("Resetting")
    send_hid_command(device, HID_COMMAND_RESET)


# region: Update Application
def check_for_self_update(major: int, minor: int, patch: int, proxy: str | None = None):
    if proxy:
        proxies = {"https": proxy}
    else:
        proxies = {}
    try:
        result = requests.get(
            "https://api.github.com/repos/mutenix-org/software-host/releases/latest",
            timeout=4,
            proxies=proxies,
        )
        if result.status_code != 200:
            _logger.error(
                "Failed to fetch latest release info, status code: %s",
                result.status_code,
            )
            return

        releases = result.json()
        latest_version = releases.get("tag_name", "v0.0.0")[1:]
        _logger.debug("Latest version: %s", latest_version)
        online_version = semver.Version.parse(latest_version)
        local_version = semver.Version(major=major, minor=minor, patch=patch)
        if online_version.compare(local_version) <= 0:
            _logger.info("Host Software is up to date")
            return

        _logger.info("Application update available, but auto update is disabled")
        webbrowser.open(releases.get("html_url"))
    except requests.RequestException as e:
        _logger.error("Failed to check for application update availability: %s", e)


# endregion



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import tracemalloc
from unittest.mock import ANY
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from mutenix.config import DeviceInfo
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import Ping
from mutenix.hid_commands import PrepareUpdate
from mutenix.hid_device import HidDevice

tracemalloc.start()


@pytest.fixture
def hid_device():
    with patch("hid.device") as MockHidDevice:
        MockHidDevice.return_value = Mock()
        return HidDevice()


@pytest.mark.asyncio
async def test_send_msg(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    assert not future.done()
    assert not hid_device._send_buffer.empty()


@pytest.mark.asyncio
async def test_register_callback(hid_device):
    callback = Mock()
    hid_device.register_callback(callback)
    assert callback in hid_device._callbacks


@pytest.mark.asyncio
async def test_ping(hid_device):
    with patch.object(hid_device, "send_msg", return_value=Mock()):
        with patch("asyncio.sleep", side_effect=asyncio.CancelledError):
            with pytest.raises(asyncio.CancelledError):
                await hid_device._ping()


@pytest.mark.asyncio
async def test_write_success(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(hid_device, "_send_report", return_value=1):
        await hid_device._write()
    assert future.done()
    assert future.result() == 1


@pytest.mark.asyncio
async def test_write_failure(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(hid_device, "_send_report", return_value=-1):
        await hid_device._write()
    assert future.done()
    assert future.exception() is not None


@pytest.mark.asyncio
async def test_write_device_disconnected(hid_device):
    msg = HidOutputMessage()
    hid_device._wait_for_device = AsyncMock(return_value=None)
    future = hid_device.send_msg(msg)
    with patch.object(
        hid_device,
        "_send_report",
        side_effect=OSError("Device disconnected"),
    ):
        await hid_device._write()
    assert future._exception is not None


@pytest.mark.asyncio
async def test_write_value_error(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    hid_device._wait_for_device = AsyncMock(return_value=None)
    with patch.object(
        hid_device,
        "_send_report",
        side_effect=ValueError("Invalid message"),
    ):
        await hid_device._write()
    assert future._exception is not None


@pytest.mark.asyncio
async def test_read_success(hid_device):
    data = bytes([0x01, 0x02, 0x03])
    future = asyncio.get_event_loop().create_future()

    async def callback(msg):
        future.set_result(msg)

    hid_device._device = Mock()
    hid_device._device.read.return_value = data

    hid_device._callbacks.append(callback)
    with patch(
        "mutenix.hid_commands.HidInputMessage.from_buffer",
        return_value=Mock(),
    ):
        await hid_device._read()
    await future


@pytest.mark.asyncio
async def test_read_device_disconnected(hid_device):
    hid_device._wait_for_device = AsyncMock()
    hid_device._device = Mock()
    hid_device._device.read.side_effect = OSError("Device disconnected")

    with patch("mutenix.hid_device._logger.error") as mock_logger:
        await hid_device._read()
        mock_logger.assert_called_with("Device disconnected: %s", ANY)

    assert hid_device._wait_for_device.called


@pytest.mark.asyncio
async def test_read_value_error(hid_device):
    hid_device._device = Mock()
    hid_device._device.read.side_effect = ValueError("Invalid message")
    with patch("mutenix.hid_device._logger.error") as mock_logger:
        await hid_device._read()
        mock_logger.assert_called_with(
            "Error reading message: %s",
            ANY,
        )


class TypeMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


@pytest.mark.asyncio
async def test_ping_sends_ping_message(hid_device):
    hid_device._last_communication = asyncio.get_event_loop().time() - 5
    with patch.object(hid_device, "send_msg") as mock_send_msg:
        await hid_device._ping()
        mock_send_msg.assert_called_once_with(TypeMatcher(Ping))


@pytest.mark.asyncio
async def test_ping_resets_last_communication(hid_device):
    initial_time = asyncio.get_event_loop().time()
    with patch("asyncio.get_event_loop", return_value=asyncio.get_event_loop()):
        with patch.object(hid_device, "send_msg"):
            await hid_device._ping()
            assert hid_device._last_ping_time >= initial_time


class RoundAboutMatcher:
    def __init__(self, expected_value, offset):
        self.expected_value = expected_value
        self.offset = offset

    def __eq__(self, other):
        return abs(other - self.expected_value) < self.offset


@pytest.mark.asyncio
async def test_ping_waits_before_sending(hid_device):
    hid_device._last_ping_time = asyncio.get_event_loop().time()
    with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
        with patch.object(hid_device, "send_msg"):
            await hid_device._ping()
            mock_sleep.assert_called_with(RoundAboutMatcher(4.5, 0.2))


# test


@pytest.mark.asyncio
async def test_send_report_success(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.return_value = 1
    result = hid_device._send_report(msg)
    assert result == 1
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_send_report_failure(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.return_value = -1
    result = hid_device._send_report(msg)
    assert result == -1
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_send_report_exception(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.side_effect = OSError("Device error")
    with pytest.raises(OSError, match="Device error"):
        hid_device._send_report(msg)
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_unregister_callback(hid_device: HidDevice):
    # Create a HidDevice instance
    device = HidDevice()

    # Create a mock callback
    callback = Mock()

    # Register the callback
    device.register_callback(callback)
    assert callback in device._callbacks

    # Unregister the callback
    device.unregister_callback(callback)
    assert callback not in device._callbacks

    # Try to unregister the callback again (should not raise an error)
    device.unregister_callback(callback)
    assert callback not in device._callbacks


@pytest.mark.asyncio
async def test_invoke_callbacks_sync(hid_device):
    callback = Mock()
    hid_device.register_callback(callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    callback.assert_called_once_with(msg)


@pytest.mark.asyncio
async def test_invoke_callbacks_async(hid_device):
    async def async_callback(msg):
        async_callback.called = True

    async_callback.called = False
    hid_device.register_callback(async_callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    await asyncio.sleep(0)  # Allow the async callback to be called
    assert async_callback.called


@pytest.mark.asyncio
async def test_invoke_callbacks_mixed(hid_device):
    callback = Mock()

    async def async_callback(msg):
        async_callback.called = True

    async_callback.called = False
    hid_device.register_callback(callback)
    hid_device.register_callback(async_callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    callback.assert_called_once_with(msg)
    await asyncio.sleep(0)  # Allow the async callback to be called
    assert async_callback.called


@pytest.mark.asyncio
async def test_search_for_device_success(hid_device):
    device_info = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
        "bus_type": 1,
    }
    with patch("hid.device"):
        with patch("hid.enumerate", return_value=[device_info]):
            device = await hid_device._search_for_device()
            assert device is not None


@pytest.mark.asyncio
async def test_search_for_device_prefer_bluetooth(hid_device):
    device_info_bt = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
        "bus_type": 2,
    }
    device_info_usb = {
        "product_id": 0x1234,
        "vendor_id": 0x5678,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
        "bus_type": 1,
    }
    with patch("hid.device") as MockHidDevice:
        with patch("hid.enumerate", return_value=[device_info_bt, device_info_usb]):
            m = Mock()
            MockHidDevice.return_value = m
            device = await hid_device._search_for_device()
            assert device is not None
            m.open.assert_called_with(product_id=0, vendor_id=0, serial_number="122345")


@pytest.mark.asyncio
async def test_search_for_device_no_device_found_device_info(hid_device):
    hid_device._device_info = [DeviceInfo()]
    with patch("hid.enumerate", return_value=[]):
        device = await hid_device._search_for_device()
        assert device is None


@pytest.mark.asyncio
async def test_search_for_device_no_device_found(hid_device):
    with patch("hid.enumerate", return_value=[]):
        device = await hid_device._search_for_device()
        assert device is None


@pytest.mark.asyncio
async def test_search_for_device_no_device_found_open_fails(hid_device):
    with patch("hid.enumerate", return_value=[]):
        with patch.object(hid_device, "_open_device_with_info", return_value=None):
            device = await hid_device._search_for_device()
            assert device is None


@pytest.mark.asyncio
async def test_search_for_device_exception(hid_device):
    with patch("hid.enumerate", side_effect=Exception("Error")):
        with patch("mutenix.hid_device._logger.debug") as mock_logger:
            device = await hid_device._search_for_device()
            assert device is None
            mock_logger.assert_called_with("Failed to get device: %s", ANY)


@pytest.mark.asyncio
async def test_wait_for_device_success(hid_device):
    with patch.object(
        hid_device,
        "_search_for_device_loop",
        new_callable=AsyncMock,
    ) as mock_search:
        mock_search.return_value = Mock()
        await hid_device._wait_for_device()
        assert mock_search.called
        assert hid_device._device is not None


@pytest.mark.asyncio
async def test_wait_for_device_failure(hid_device):
    with patch.object(
        hid_device,
        "_search_for_device_loop",
        new_callable=AsyncMock,
    ) as mock_search:
        mock_search.return_value = None
        await hid_device._wait_for_device()
        assert mock_search.called
        assert hid_device._device is None


@pytest.mark.asyncio
async def test_open_device_with_info_bt_success(hid_device):
    device_info = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
    }
    with patch("hid.device") as MockHidDevice:
        mock_device = Mock()
        MockHidDevice.return_value = mock_device
        device = hid_device._open_device_with_info(device_info)
        assert device is not None
        mock_device.open.assert_called_once_with(
            product_id=0,
            vendor_id=0,
            serial_number="122345",
        )


@pytest.mark.asyncio
async def test_open_device_with_info_bt_failure(hid_device):
    device_info = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
        "bus_type": 2,
    }
    with patch("hid.device") as MockHidDevice:
        mock_device = Mock()
        MockHidDevice.return_value = mock_device
        mock_device.open.side_effect = Exception("BT Connection error")
        with patch("mutenix.hid_device._logger.info") as mock_logger:
            device = hid_device._open_device_with_info(device_info)
            assert device is None
            mock_logger.assert_called_once_with(
                "Could not open device by serial number %s",
                ANY,
            )


@pytest.mark.asyncio
async def test_open_device_with_info_usb_success(hid_device):
    device_info = {
        "product_id": 0x1234,
        "vendor_id": 0x5678,
        "serial_number": "122345",
    }
    with patch("hid.device") as MockHidDevice:
        mock_device = Mock()
        MockHidDevice.return_value = mock_device
        device = hid_device._open_device_with_info(device_info)
        assert device is not None
        mock_device.open.assert_called_once_with(
            product_id=0x1234,
            vendor_id=0x5678,
            serial_number="122345",
        )


@pytest.mark.asyncio
async def test_open_device_with_info_usb_failure(hid_device):
    device_info = {
        "product_id": 0x1234,
        "vendor_id": 0x5678,
        "serial_number": "122345",
    }
    with patch("hid.device") as MockHidDevice:
        mock_device = Mock()
        MockHidDevice.return_value = mock_device
        mock_device.open.side_effect = Exception("USB Connection error")
        with patch("mutenix.hid_device._logger.info") as mock_logger:
            device = hid_device._open_device_with_info(device_info)
            assert device is None
            mock_logger.assert_called_once_with(
                "Could not open HID Connection (%s)",
                ANY,
            )


@pytest.mark.asyncio
async def test_send_report_device_not_connected(hid_device):
    msg = PrepareUpdate()

    hid_device._device = None
    with pytest.raises(ValueError, match="Device not connected"):
        hid_device._send_report(msg)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import logging
from typing import Callable

import hid
from mutenix.config import DeviceInfo
from mutenix.hid_commands import HidCommand
from mutenix.hid_commands import HidInputMessage
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import Ping
from mutenix.utils import block_parallel
from mutenix.utils import rate_limited_logger
from mutenix.utils import run_loop
from mutenix.utils import run_till_some_loop

_logger = logging.getLogger(__name__)


class HidDevice:
    """Handles the HID connection to the device.
    Providing async read and write loops for incoming and outgoing messages.
    """

    def __init__(self, device_identifications: list[DeviceInfo] | None = None):
        self._device_info = device_identifications or []
        self._device: hid.device | None = None
        self._callbacks: list[Callable[[HidInputMessage], None]] = []
        self._send_buffer: asyncio.Queue[tuple[HidOutputMessage, asyncio.Future]] = (
            asyncio.Queue()
        )
        self._last_communication: float = 0
        self._last_ping_time: float = 0
        self._waiting_for_device: bool = False
        self._run: bool = True

    def __del__(self):
        if self._device:
            self._device.close()

    @block_parallel
    async def _wait_for_device(self):
        _logger.info(
            "Looking for device with",
        )
        self._device = None
        self._device = await self._search_for_device_loop()

    def _open_device_with_info(self, device_info: DeviceInfo):
        device = hid.device()
        if device_info["product_id"] == 0 and device_info["vendor_id"] == 0:
            try:
                device.open(
                    product_id=0,
                    vendor_id=0,
                    serial_number=device_info["serial_number"],
                )
                return device
            except Exception as e:
                _logger.info("Could not open device by serial number %s", e)
        else:
            try:
                device.open(
                    product_id=device_info["product_id"],
                    vendor_id=device_info["vendor_id"],
                    serial_number=device_info["serial_number"],
                )
                return device
            except Exception as e:
                _logger.info("Could not open HID Connection (%s)", e)
        return None

    async def _search_for_device(self):
        try:

            def find_device():
                devices = []
                for device in sorted(
                    hid.enumerate(),
                    key=lambda x: x["bus_type"],
                    reverse=True,
                ):
                    if "mutenix" in device["product_string"].lower():
                        _logger.info("Device found %s", device)
                        devices.append(device)
                return devices

            if not self._device_info or len(self._device_info) == 0:
                available_devices = find_device()
            else:
                available_devices = list(
                    map(lambda x: x.model_dump(), self._device_info),
                )
            if len(available_devices) == 0:
                _logger.error("No device available, no config")
                await asyncio.sleep(0)
                return None
            _logger.debug("Looking for device with %s", available_devices)
            device = hid.device()
            # We are sorting the devices by vendor_id to make sure we try to open BT device first
            for device_info in sorted(available_devices, key=lambda x: x["vendor_id"]):
                if device := self._open_device_with_info(device_info):
                    break
                _logger.debug("Device not found %s", device_info)
            else:
                return None
            _logger.info("Device found %s", device)
            device.set_nonblocking(1)
            return device
        except Exception as e:
            _logger.debug("Failed to get device: %s", e)
            return None

    def _send_report(self, data: HidCommand):
        buffer = bytes([data.REPORT_ID]) + data.to_buffer()
        buffer = bytes(buffer)
        if not self._device:
            raise ValueError("Device not connected")
        return self._device.write(buffer)

    def send_msg(self, msg: HidOutputMessage):
        """
        Sends a HID output message asynchronously.

        Args:
            msg (HidOutputMessage): The HID output message to be sent.

        Returns:
            asyncio.Future: A future that will be set when the message is processed.
        """
        future = asyncio.get_event_loop().create_future()
        self._send_buffer.put_nowait((msg, future))
        _logger.debug("Put message")
        return future

    def register_callback(self, callback):
        if callback not in self._callbacks:
            self._callbacks.append(callback)

    def unregister_callback(self, callback):
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _invoke_callbacks(self, msg):
        for callback in self._callbacks:
            if asyncio.iscoroutinefunction(callback):
                asyncio.create_task(callback(msg))
            else:
                callback(msg)

    @rate_limited_logger(_logger, limit=3, interval=10)
    def _log_failed_to_send(self, *args, **kwargs):
        _logger.error(*args, **kwargs)

    async def _read(self):
        try:
            buffer: list = self._device.read(64)
            if buffer and len(buffer):
                msg = HidInputMessage.from_buffer(buffer)
                self._invoke_callbacks(msg)
            else:
                await asyncio.sleep(0.1)  # pragma: no cover
        except OSError as e:  # Device disconnected
            _logger.error("Device disconnected: %s", e)
            await self._wait_for_device()
            _logger.info("Device reconnected")
        except Exception as e:
            _logger.error("Error reading message: %s", e)

    async def _write(self):
        try:
            msg, future = await self._send_buffer.get()
            _logger.debug("Sending message: %s", msg)
            result = self._send_report(msg)
            if result < 0:
                self._log_failed_to_send("Failed to send message: %s", msg)
                future.set_exception(Exception("Failed to send message"))
                return
            self._last_communication = asyncio.get_event_loop().time()
            if not future.cancelled():
                future.set_result(result)
            self._send_buffer.task_done()
        except OSError as e:  # Device disconnected
            _logger.error("Device disconnected: %s", e)
            future.set_exception(e)
            await self._wait_for_device()
        except ValueError as e:
            _logger.error("Error sending message: %s", e)
            future.set_exception(e)
            await self._wait_for_device()

    async def _ping(self):
        """
        Sends a ping message to the HID device.
        """
        await asyncio.sleep(
            self._last_ping_time + 4.5 - asyncio.get_event_loop().time(),
        )
        if asyncio.get_event_loop().time() - self._last_ping_time > 4.5:
            _logger.debug("Sending ping")
            msg = Ping()
            future = self.send_msg(msg)
            try:
                self._last_ping_time = asyncio.get_event_loop().time()
                await future
                _logger.debug("ping finally sent")
            except Exception as e:
                self._log_failed_to_send("Failed to send ping: %s", e)
            self._last_ping_time = asyncio.get_event_loop().time()

    async def _process(self):  # pragma: no cover
        await self._wait_for_device()
        await asyncio.gather(self._read_loop(), self._write_loop(), self._ping_loop())

    async def process(self):  # pragma: no cover
        """
        Processes the HID device by running the internal process loop.

        Returns:
            None
        """
        await self._process_loop()

    async def stop(self):  # pragma: no cover
        """
        Stops the HID device by setting the internal run flag to False.

        Signals the processing loops to stop.
        """
        self._run = False

    # create the run loops
    _read_loop = run_loop(_read)
    _write_loop = run_loop(_write)
    _ping_loop = run_loop(_ping)
    _process_loop = run_loop(_process)
    _search_for_device_loop = run_till_some_loop(sleep_time=1)(_search_for_device)

    @property
    def raw(self):  # pragma: no cover
        return self._device

    async def wait_for_device(self):  # pragma: no cover
        await self._wait_for_device()

    @property
    def connected(self) -> bool:  # pragma: no cover
        return self._device is not None



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import logging
from typing import Callable

import websockets
from mutenix.teams_messages import ClientMessage
from mutenix.teams_messages import ServerMessage
from mutenix.utils import block_parallel
from mutenix.utils import rate_limited_logger
from mutenix.utils import run_loop

_logger = logging.getLogger(__name__)


@rate_limited_logger(_logger, limit=3, interval=10)
def _log_failed_to_connect(*args, **kwargs):
    _logger.error(*args, **kwargs)


class Identifier:
    """Identifies the device to the Teams Websocket server.
    The token is used to authenticate the device with the server. It is received once the user
    allows the device to connect to Teams."""

    def __init__(self, manufacturer, device, app, app_version, token=""):
        self.protocol_version = "2.0.0"
        self.manufacturer = manufacturer
        self.device = device
        self.app = app
        self.app_version = app_version
        self.token = token


class WebSocketClient:
    """Handles the WebSocket connection to Teams."""

    RETRY_INTERVAL = 0.25

    def __init__(self, uri: str, identifier: Identifier):
        self._uri = uri
        self._connection = None
        self._send_queue: asyncio.Queue[tuple[ClientMessage, asyncio.Future]] = (
            asyncio.Queue()
        )
        self._callback: Callable[[ServerMessage], None] | None = None
        params = (
            f"?protocol-version={identifier.protocol_version}"
            f"&manufacturer={identifier.manufacturer}"
            f"&device={identifier.device}"
            f"&app={identifier.app}"
            f"&app-version={identifier.app_version}"
            f"&token={identifier.token}"
        )
        self._uri += params
        self._connecting = False
        self._run = True
        self._sent_something = True

    @block_parallel
    async def _connect(self):
        while self._run:
            self._connection = None
            connection = await self._do_connect()
            if not connection:
                await asyncio.sleep(self.RETRY_INTERVAL)
            else:
                self._connection = connection
                break

    async def _do_connect(self):
        try:
            connection = await websockets.connect(self._uri)
            _logger.info("Connected to WebSocket server at %s", self._uri)
            return connection
        except Exception as e:
            _log_failed_to_connect(
                "Failed to connect to WebSocket server: %s: %s",
                type(e).__name__,
                e,
            )
            return None

    def send_message(self, message: ClientMessage):
        future = asyncio.get_event_loop().create_future()
        self._send_queue.put_nowait((message, future))
        return future

    def register_callback(self, callback: Callable[[ServerMessage], None]):
        self._callback = callback

    async def _send(self):
        try:
            queue_element = self._send_queue.get_nowait()
            message, future = queue_element
        except asyncio.QueueEmpty:
            if self._sent_something:
                self._sent_something = False
                _logger.debug("Send queue empty")
            await asyncio.sleep(0.2)
            return
        try:
            if isinstance(message, ClientMessage):
                msg = message.model_dump_json(by_alias=True)
                self._sent_something = True
            else:
                future.set_exception(
                    TypeError("Expected message to be an instance of ClientMessage"),
                )
                return
            await self._connection.send(msg)
            if not future.done():
                future.set_result(True)
        except Exception as e:
            future.set_exception(e)
            await self._connect()
        finally:
            self._send_queue.task_done()

    async def _receive(self):
        try:
            async with asyncio.timeout(1):
                msg = await self._connection.recv()
                _logger.debug("Received message: %s", msg)
                message = ServerMessage.model_validate_json(msg)
                if message:
                    _logger.debug("Decoded message: %s", message)
                if self._callback:
                    if asyncio.iscoroutinefunction(self._callback):
                        asyncio.create_task(self._callback(message))
                    else:
                        self._callback(message)
        except asyncio.TimeoutError:
            pass
        except Exception as e:
            _logger.error("Error receiving message: %s", e)
            await self._connect()

    async def process(self):
        while self._run:
            asyncio.current_task().set_name("WebSocketClient.process")
            try:
                await self._connect()
                await asyncio.gather(self._send_loop(), self._receive_loop())
            except asyncio.CancelledError as e:
                _logger.info("shutting down due to cancel: %s", e)
                await self.stop()

    async def stop(self):
        self._run = False
        if self._connection:
            await self._connection.close()
        while not self._send_queue.empty():
            _, future = self._send_queue.get_nowait()
            future.set_exception(RuntimeError("WebSocketClient is stopping"))
            self._send_queue.task_done()

    @property
    def connected(self) -> bool:  # pragma: no cover
        return self._connection is not None

    _receive_loop = run_loop(_receive)
    _send_loop = run_loop(_send)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import logging
import pathlib
import sys
import tempfile

import win32api  # type: ignore
import win32con  # type: ignore
import win32event  # type: ignore
import win32gui  # type: ignore
from pywinauto.findwindows import find_windows  # type: ignore
from winerror import ERROR_ALREADY_EXISTS  # type: ignore

_logger = logging.getLogger(__name__)


def bring_teams_to_foreground() -> None:  # pragma: no cover
    """
    Bring the Microsoft Teams window to the foreground.

    This function attempts to bring the Microsoft Teams application window to the foreground
    on different operating systems (Windows, macOS, and Linux). It uses platform-specific
    methods to achieve this.

    On Windows, it uses the `win32gui` and `win32con` modules to minimize and restore the window.
    Note: This function will not be coverable due to its OS dependencies.
    """
    try:
        _logger.debug("Finding Microsoft Teams window")
        window_id = find_windows(title_re=".*Teams.*")
        _logger.debug("Window ID: %s", window_id)
        for w in window_id:
            _logger.debug("Minimizing and restoring window %s", w)
            win32gui.ShowWindow(w, win32con.SW_MINIMIZE)
            win32gui.ShowWindow(w, win32con.SW_RESTORE)
            win32gui.ShowWindow(w, win32con.SW_SHOWMAXIMIZED)
            win32gui.ShowWindow(w, win32con.SW_SHOWNORMAL)
            win32gui.SetActiveWindow(w)
    except Exception as e:
        _logger.warning(
            "Could not bring Microsoft Teams window to the foreground, %s",
            e,
        )


def ensure_process_run_once(
    lockfile_path: pathlib.Path = pathlib.Path(tempfile.gettempdir()),
):
    def outerwrapper(func):
        def wrapper(*args, **kwargs):
            mutex = win32event.CreateMutex(None, False, "mutenix_instance")
            last_error = win32api.GetLastError()

            if last_error == ERROR_ALREADY_EXISTS:
                print("App instance already running")
                _logger.error("Another instance of the application is already running.")
                sys.exit(1)
            try:
                result = func(*args, **kwargs)
            finally:
                win32event.ReleaseMutex(mutex)
            return result

        return wrapper

    return outerwrapper



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import logging
import shlex
import subprocess
import time
from collections import defaultdict
from typing import Callable

import requests
from mutenix.config import ActionEnum
from mutenix.config import ButtonAction
from mutenix.config import Config
from mutenix.config import LedStatusSource
from mutenix.config import load_config
from mutenix.config import save_config
from mutenix.config import WebhookAction
from mutenix.hid_commands import LedColor
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.hid_commands import UpdateConfig
from mutenix.hid_commands import VersionInfo
from mutenix.hid_device import HidDevice
from mutenix.teams_messages import ClientMessage
from mutenix.teams_messages import ClientMessageParameter
from mutenix.teams_messages import ClientMessageParameterType
from mutenix.teams_messages import MeetingAction
from mutenix.teams_messages import ServerMessage
from mutenix.updates import check_for_device_update
from mutenix.updates import perform_upgrade_with_file
from mutenix.utils import bring_teams_to_foreground
from mutenix.utils import run_loop
from mutenix.virtual_macropad import VirtualMacropad
from mutenix.websocket_client import Identifier
from mutenix.websocket_client import WebSocketClient

try:
    from pynput.keyboard import Controller
    from pynput.keyboard import Key
    from pynput.mouse import Button
    from pynput.mouse import Controller as MouseController
except ImportError:  # pragma: no cover
    Controller = None
    Key = None
    Button = None
    MouseController = None

_logger = logging.getLogger(__name__)


class Macropad:
    """The main logic for the Macropad."""

    def __init__(self, config: Config):
        self._run = True
        self._version_seen = None
        self._last_status_check: defaultdict[int, int] = defaultdict(int)
        self._config = config
        self._last_led_update: dict[int, SetLed] = {}
        self._setup()
        self._current_state: ServerMessage | None = None
        self._setup_buttons()
        self._checktime = time.time()

    def _setup(self):
        self._device = HidDevice(self._config.device_identifications)
        token = self._config.teams_token
        self._websocket = WebSocketClient(
            "ws://127.0.0.1:8124",
            Identifier(
                manufacturer="test",
                device="test",
                app="test",
                app_version="1.0.0",
                token=token,
            ),
        )
        self._virtual_macropad = VirtualMacropad(
            self._config.virtual_keypad.bind_address,
            self._config.virtual_keypad.bind_port,
        )
        self._virtual_macropad.update_config(self._config)
        self._websocket.register_callback(self._teams_callback)
        self._device.register_callback(self._hid_callback)
        self._virtual_macropad.register_callback(self._hid_callback)

    def _setup_buttons(self):
        self._tap_actions = {entry.button_id: entry for entry in self._config.actions}
        self._longpress_actions = {
            entry.button_id: entry for entry in self._config.longpress_action
        }

    def _perform_webhook(self, extra: WebhookAction):
        try:
            result = requests.request(
                extra.method,
                extra.url,
                json=extra.data,
                headers={str(key): str(value) for key, value in extra.headers.items()},
            )
            _logger.info("Webhook result: %i %s", result.status_code, result.text)
        except Exception as e:
            _logger.warn("Webhook resulted in an exeption %s", e)

    def _keypress(self, extra):
        if not Controller:
            _logger.error("pynput not supported, cannot send keypress")
            return
        if isinstance(extra, list):  # pragma: no cover
            for sequence in extra:
                self._keypress(sequence)
            return

        keyboard = Controller()
        _logger.debug("Keypress: %s", extra)
        if "key" in extra:

            def do_key(*keys):
                if len(keys) == 1:
                    keyboard.tap(keys[0])
                else:
                    with keyboard.pressed(keys[0]):
                        do_key(*keys[1:])

            try:
                do_key(
                    *map(lambda x: getattr(Key, x), extra.get("modifiers", [])),
                    extra["key"]
                    if len(extra["key"]) == 1
                    else getattr(Key, extra["key"]),
                )
            except AttributeError:
                _logger.warning("Key not found")
        elif "string" in extra:
            keyboard.type(extra["string"])

    def _mousemove(self, extra):
        if not MouseController:
            _logger.error("pynput not supported, cannot send mousemove")
            return
        if isinstance(extra, list):  # pragma: no cover
            for sequence in extra:
                self._mousemove(sequence)
            return
        mouse = MouseController()
        action = extra.get("action", "move")
        match action:
            case "move":
                mouse.move(extra.get("x", 0), extra.get("y", 0))
            case "set":
                mouse.position = (extra.get("x", 0), extra.get("y", 0))
            case "click":
                mouse.click(getattr(Button, extra["button"]), extra.get("count", 1))
            case "press":
                mouse.press(getattr(Button, extra["button"]))
            case "release":
                mouse.release(getattr(Button, extra["button"]))

    def _do_run_command(self, command):
        _logger.debug("Running command: %s", command)
        result = subprocess.run(
            shlex.split(command),
            capture_output=True,
            text=True,
        )
        _logger.debug("Command output: %s", result.stdout)
        _logger.debug("Command error: %s", result.stderr)
        _logger.debug("Command return code: %s", result.returncode)

    def _run_command(self, extra):  # pragma: no cover
        for command in extra:
            try:
                asyncio.create_task(asyncio.to_thread(self._do_run_command, command))
            except Exception as e:
                _logger.error("Error running command: %s", e)

    async def _send_status(self, status: Status):
        _logger.info(
            "Button %s, Triggered: %s, Longpress: %s",
            status.button,
            status.triggered,
            status.longpressed,
        )
        _logger.debug("Status: %s", status)
        action: None | ButtonAction = None
        mapped_action: Callable | None | MeetingAction = None

        action_map: dict[ActionEnum, Callable] = {
            ActionEnum.ACTIVATE_TEAMS: lambda _: bring_teams_to_foreground(),
            ActionEnum.CMD: self._run_command,
            ActionEnum.WEBHOOK: self._perform_webhook,
            ActionEnum.KEYPRESS: self._keypress,
            ActionEnum.MOUSE: self._mousemove,
        }

        if status.triggered:
            if not status.released:
                return
            if not status.longpressed and status.button in self._tap_actions:
                action = self._tap_actions.get(status.button, None)
            elif status.longpressed and status.button in self._longpress_actions:
                action = self._longpress_actions.get(status.button, None)
            if not action:
                return
            if isinstance(action.action, MeetingAction):
                if action.action == MeetingAction.React and isinstance(
                    action.extra,
                    ClientMessageParameterType,
                ):
                    client_message = ClientMessage.create(
                        action=MeetingAction.React,
                    )
                    client_message.parameters = ClientMessageParameter(
                        type_=action.extra,
                    )
                else:
                    client_message = ClientMessage.create(action=action.action)
                await self._websocket.send_message(client_message)
            else:
                mapped_action = action_map.get(action.action, None)
                if callable(mapped_action):
                    mapped_action(action.extra)  # pragma: no cover

    async def _process_version_info(self, version_info: VersionInfo):
        if self._version_seen != version_info.version:
            _logger.info(version_info)
            self._version_seen = version_info.version
            if self._config.auto_update:
                check_for_device_update(
                    self._device.raw,
                    version_info,
                    self._config.proxy,
                )
        else:
            _logger.debug(version_info)
        await self._update_device_status(force=True)

    async def _hid_callback(self, msg):
        if isinstance(msg, Status):
            await self._send_status(msg)
        elif isinstance(msg, VersionInfo):
            await self._process_version_info(msg)

    async def _teams_callback(self, msg: ServerMessage):
        _logger.debug("Teams message: %s", msg)
        if msg.meeting_update:
            self._current_state = msg
        if msg.token_refresh:
            self._config.teams_token = msg.token_refresh
            save_config(self._config)
        await self._update_device_status()

    def _map_led_color(self, color):
        color = color.upper()
        if not hasattr(LedColor, color):
            return LedColor.GREEN
        return getattr(LedColor, color)

    async def _update_device_status(self, force=False):
        msg = self._current_state
        msgs = {}

        for ledstatus in self._config.leds:
            if ledstatus.source == LedStatusSource.TEAMS:
                color = "black"
                if (
                    msg
                    and msg.meeting_update
                    and msg.meeting_update.meeting_state
                    and msg.meeting_update.meeting_state.is_in_meeting
                ):
                    mapped_state = getattr(
                        msg.meeting_update.meeting_state,
                        ledstatus.extra.replace("-", "_").lower(),
                    )
                    color = ledstatus.color_on if mapped_state else ledstatus.color_off
                    msgs[ledstatus.button_id] = SetLed(
                        ledstatus.button_id,
                        self._map_led_color(color),
                    )
                else:
                    msgs[ledstatus.button_id] = SetLed(
                        ledstatus.button_id,
                        self._map_led_color("black"),
                    )
            elif ledstatus.source == LedStatusSource.CMD:
                if (
                    self._last_status_check[ledstatus.button_id] + ledstatus.interval
                    > time.time()
                ):
                    continue
                if ledstatus.read_result:
                    async with asyncio.timeout(ledstatus.timeout):
                        result = await asyncio.to_thread(
                            subprocess.check_output,
                            ledstatus.extra,
                        )
                        msgs[ledstatus.button_id] = SetLed(
                            ledstatus.button_id,
                            self._map_led_color(result.strip()),
                        )
                else:
                    async with asyncio.timeout(ledstatus.timeout):
                        result = await asyncio.to_thread(
                            subprocess.check_call,
                            ledstatus.extra,
                        )
                        msgs[ledstatus.button_id] = SetLed(
                            ledstatus.button_id,
                            self._map_led_color(
                                ledstatus.color_on
                                if result == 0
                                else ledstatus.color_off,
                            ),
                        )
                self._last_status_check[ledstatus.button_id] = time.time()
            elif ledstatus.source == LedStatusSource.WEBHOOK:
                color = self._virtual_macropad.get_led_status(ledstatus.button_id)
                msgs[ledstatus.button_id] = SetLed(
                    ledstatus.button_id,
                    self._map_led_color(color),
                )

        for key, message in msgs.items():
            try:
                if not force and (
                    key in self._last_led_update
                    and self._last_led_update[key] == message
                ):
                    continue
                _logger.debug(
                    f"Sending message: {message}, prev: {self._last_led_update.get(key, None)}",
                )
                self._device.send_msg(message)
                await self._virtual_macropad.send_msg(message)
                self._last_led_update[key] = message
            except Exception as e:
                _logger.exception(e)

    async def _do_check_status(self):
        from mutenix.tray_icon import my_icon

        await self._update_device_status()
        await asyncio.sleep(0.1)
        if int(time.time() - self._checktime) > 10:
            try:
                if my_icon:  # pragma: no cover
                    my_icon.update_menu()
                self._checktime = time.time()
            except Exception as e:  # pragma: no cover
                _logger.error("Error updating tray icon: %s", e)
                print(e)

    _check_status = run_loop(_do_check_status)

    async def process(self):
        """Starts the process loop for the device and the WebSocket connection."""
        try:
            await asyncio.gather(
                self._device.process(),
                self._websocket.process(),
                self._virtual_macropad.process(),
                self._check_status(),
            )
        except Exception as e:
            _logger.error("Error in Macropad process: %s", e)

    async def manual_update(self, update_file):
        """Manually update the device with a given file."""
        await self._device.wait_for_device()
        with open(update_file, "rb") as f:
            perform_upgrade_with_file(self._device.raw, f)

    async def stop(self):
        """Stops the device and WebSocket connection."""
        self._run = False
        await self._device.stop()
        _logger.info("Device stopped")
        await self._websocket.stop()
        _logger.info("Websocket stopped")
        await self._virtual_macropad.stop()
        _logger.info("Virtual Device stopped")

    @property
    def virtual_keypad_address(self):  # pragma: no cover
        return self._config.virtual_keypad.bind_address

    @property
    def virtual_keypad_port(self):  # pragma: no cover
        return self._config.virtual_keypad.bind_port

    def activate_serial_console(self):
        message = UpdateConfig()
        message.activate_serial_console(True)
        self._device.send_msg(message)

    def deactivate_serial_console(self):
        message = UpdateConfig()
        message.activate_serial_console(False)
        self._device.send_msg(message)

    def activate_filesystem(self):
        message = UpdateConfig()
        message.activate_filesystem(True)
        self._device.send_msg(message)

    @property
    def teams_connected(self) -> bool:  # pragma: no cover
        return self._websocket.connected

    @property
    def device_connected(self) -> bool:  # pragma: no cover
        return self._device.connected

    def reload_config(self):
        _logger.info("Reloading config")
        self._config = load_config()
        self._setup_buttons()
        self._update_device_status(force=True)
        self._virtual_macropad.update_config(self._config)
        _logger.info("Config reloaded")
